//
//  ScrunchyView.m
//  seamus
//
//  Created by Wim Lewis on 23Aug07.
//  Copyright 2007 Wim Lewis.
//  You may distribute and/or modify this code under the terms of the GNU General Public License as published by the Free Software Foundation, version 2, or (at your option) any later version.
//

#import "ScrunchyView.h"
#import "Seamer.h"
#import <CoreServices/CoreServices.h>

@implementation ScrunchyView

- (id)initWithFrame:(NSRect)frame {
    self = [super initWithFrame:frame];
    if (self) {
        // Initialization code here.
    }
    return self;
}

- (void)drawRect:(NSRect)rect
{
    [[NSColor blackColor] set];
    NSRectFill(rect);
    
    if (sourceImage && seamer && !scrunchedImage)
        [self reduce];
    
    if (!scrunchedImage)
        return;
    
    int cols = seamer->cols;
    NSRect bs = [self bounds];
    
    bs.size.width = bs.size.width * [scrunchedImageRep pixelsWide] / cols;
    
    [scrunchedImage drawInRect:bs fromRect:(NSRect){{0,0},[scrunchedImage size]} operation:NSCompositeSourceOver fraction:1.0];
}

- (void)clearBuffers
{
    [scrunchedImageRep release];
    scrunchedImageRep = nil;
    [scrunchedImage release];
    scrunchedImage = nil;
    if (scrunchBuffer) {
        free(scrunchBuffer);
        scrunchBuffer = NULL;
    }
}

- (void)setImage:(NSImage *)source;
{
    [sourceImage release];
    sourceImage = nil;
    [seamer release];
    seamer = nil;
    [self clearBuffers];
    
    for(int ix = 0; ix < [[source representations] count]; ix ++) {
        NSBitmapImageRep *r = [[source representations] objectAtIndex:ix];
        if ([r isKindOfClass:[NSBitmapImageRep class]]) {
            sourceImage = [r retain];
            break;
        }
    }
    
    [self setNeedsDisplay:YES];
}

- (void)setSeamer:(Seamer *)sr;
{
    [seamer autorelease];
    seamer = [sr retain];
    [self clearBuffers];
    [self setNeedsDisplay:YES];
}

- (void)setReduction:(int)reduce
{
    if (reduce != reduction) {
        reduction = reduce;
        [scrunchedImageRep release];
        scrunchedImageRep = nil;
        [scrunchedImage release];
        scrunchedImage = nil;
        [self setNeedsDisplay:YES];
    }
}

/*
 This method does the actual image reduction. We use a scrunchBuffer that's the same size as the original image's buffer, and implicitly specify per-scanline padding at the end to make everything work out.
 */
- (void)reduce
{
    [scrunchedImage release];
    scrunchedImage = nil;
    [scrunchedImageRep release];
    scrunchedImageRep = nil;
    
    if (!sourceImage || !seamer)
        return;
    
    int rows = seamer->rows;
    int cols = seamer->cols;

    if (reduction > seamer->seamcount)
        reduction = seamer->seamcount;  // can't reduce by more pixels than we have seams
    
    // Grab information about the image
    const unsigned char *fullBuffer = [sourceImage bitmapData];
    int bitspp = [sourceImage bitsPerPixel];  // bits per pixel
    assert(bitspp % 8 == 0);
    int bytespp = bitspp / 8;                 // bytes per pixel
    int bpr = [sourceImage bytesPerRow];      // bytes per row (scanline)
    
    if (!scrunchBuffer)
        scrunchBuffer = malloc(bpr * rows);
    
    for(int row = 0; row < rows; row ++) {
        const unsigned char *sourcePtr = fullBuffer + (row * bpr);
        unsigned char *destPtr = scrunchBuffer + (row * bpr);
        int sourceCol = 0;
        int skipIndex = 0;
        short *skips = [seamer makeSkipList:row count:reduction];
        
        // Copy pixels from *sourcePtr (in fullBuffer) to *destPtr (in scrunchBuffer), skipping over any columns listed in the skips[] array.
        while(sourceCol < cols) {
            while(skipIndex < reduction && skips[skipIndex] == sourceCol) {
                skipIndex ++;
                sourceCol ++;
                sourcePtr += bytespp;
            }
            if (!(sourceCol < cols))
                break;
            
            for(int s = 0; s < bytespp; s++)
                *destPtr++ = *sourcePtr++;
            sourceCol ++;
        }
        
        free(skips);
    }
    
    // Create a bitmap image pointing to our scrunchBuffer and store it in an instance variable for later display.
    
    NSBitmapImageRep *scrunchRep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:&scrunchBuffer
                                                                           pixelsWide:cols-reduction pixelsHigh:rows
                                                                        bitsPerSample:[sourceImage bitsPerSample]
                                                                      samplesPerPixel:[sourceImage samplesPerPixel]
                                                                             hasAlpha:[sourceImage hasAlpha]
                                                                             isPlanar:NO
                                                                       colorSpaceName:[sourceImage colorSpaceName]
                                                                         bitmapFormat:[sourceImage bitmapFormat]
                                                                          bytesPerRow:bpr
                                                                         bitsPerPixel:bitspp];
    NSImage *outimg = [[NSImage alloc] initWithSize:(NSSize){cols-reduction, rows}];
    [outimg addRepresentation:scrunchRep];
    scrunchedImageRep = scrunchRep;
    scrunchedImage = outimg;
}

- (void)setTestPoint:(NSPoint)p
{
    // Convert mouse coordinates to image coordinates and figure out how much reduction the user is asking for.
    
    int cols = seamer->cols;
    NSRect bs = [self bounds];
    
    p.x = (p.x - bs.origin.x) * ( cols / bs.size.width );
    
    int newCol = p.x;
    if (newCol < (cols - seamer->seamcount))
        newCol = cols - seamer->seamcount;
    if (newCol > cols)
        newCol = cols;
    
    [self setReduction:cols - newCol];
}

- (void)mouseDown:(NSEvent *)theEvent;
{
    NSPoint p = [self convertPoint:[theEvent locationInWindow] fromView:nil];
    [self setTestPoint:p];
}

- (void)mouseDragged:(NSEvent *)theEvent;
{
    NSPoint p = [self convertPoint:[theEvent locationInWindow] fromView:nil];
    [self setTestPoint:p];
}

static struct {
    NSBitmapImageFileType filetype;
    NSString *uti;
} filetypes[] = {
    { NSTIFFFileType, @"public.tiff" },
    { NSBMPFileType, @"com.microsoft.bmp" },
    { NSGIFFileType, @"com.compuserve.gif" },
    { NSJPEGFileType, @"public.jpeg" },
    { NSPNGFileType, @"public.png" },
    { NSJPEG2000FileType, @"public.jpeg-2000" },
    { 0, nil }
};

- (IBAction)saveTo:sender;
{
    if (sourceImage && seamer && !scrunchedImage)
        [self reduce];
    
    if (!scrunchedImage) {
        NSBeep();
        return;
    }
    
    // Set up the popup button items & tags
    if (!setupPopup) {
        [saveTypePopup removeAllItems];
        for(int ftype = 0; filetypes[ftype].uti != nil; ftype ++) {
            CFStringRef title = UTTypeCopyDescription((CFStringRef)filetypes[ftype].uti);
            NSMenuItem *item = [[saveTypePopup menu] addItemWithTitle:(NSString *)title action:NULL keyEquivalent:@""];
            CFRelease(title);
            [item setTag:filetypes[ftype].filetype];
            [item setRepresentedObject:filetypes[ftype].uti];
        }
        [saveTypePopup selectItemWithTag:NSPNGFileType];
        setupPopup = YES;
    }
    
    savePanel = [NSSavePanel savePanel];
    [savePanel setAccessoryView:saveAccessory];
    [saveTypePopup synchronizeTitleAndSelectedItem];
    [self filetypeChanged:nil];
    if ([savePanel runModal] == NSFileHandlingPanelOKButton) {
        NSError *err = nil;
        BOOL ok = [[scrunchedImageRep representationUsingType:[saveTypePopup selectedTag] properties:nil] writeToURL:[savePanel URL] options:0 error:&err];
        if (!ok)
            [self presentError:err];
    }
    [savePanel setAccessoryView:nil];
    savePanel = nil;
}

- (IBAction)filetypeChanged:sender
{
    CFStringRef extension = UTTypeCopyPreferredTagWithClass((void *)[[saveTypePopup selectedItem] representedObject], kUTTagClassFilenameExtension);
    if (extension) {
        [savePanel setRequiredFileType:(NSString *)extension];
        CFRelease(extension);
    }
}

@end

