//
//  Seamer.m
//  seamus
//
//  Created by Wim Lewis on 22Aug07.
//  Copyright 2007 Wim Lewis.
//  You may distribute and/or modify this code under the terms of the GNU General Public License as published by the Free Software Foundation, version 2, or (at your option) any later version.
//

#import "Seamer.h"
#import "SeamController.h"

#define HAVE_BOOL 1
#include <pam.h>


@implementation Seamer

- initWithImage:(NSImage *)i delegate:(SeamController *)c;
{
    [super init];
    
    int ix;
    for(ix = 0; ix < [[i representations] count]; ix ++) {
        NSBitmapImageRep *r = [[i representations] objectAtIndex:ix];
        if ([r isKindOfClass:[NSBitmapImageRep class]]) {
            baseImage = [r retain];
            width = [r pixelsWide];
            height = [r pixelsHigh];
            break;
        }
    }
    
    nonretained_delegate = c;
    
    energyFunction = NULL;
    
    pipeline = nil;
    [self setEnergyFunction:@"pamedge"];
    
    return self;
}

- (void)dealloc
{
    nonretained_delegate = nil;
    [self clearSeams];
    if(energyFunction) {
        free(energyFunction);
        energyFunction = NULL;
    }
    [pipeline release];
    [baseImage release];
    [super dealloc];
}

- (void)clearSeams
{
    if (last_upmap) {
        for(int row = 0; row < rows; row ++)
            free(last_upmap[row]);
        free(last_upmap);
        last_upmap = NULL;
    }
    
    if (seams) {
        for(int seam = 0; seam < seamcount; seam ++)
            free(seams[seam]);
        free(seams);
        seams = NULL;
        seamcount = 0;
    }
}

// Sets the middle part of the netpbm pipeline which is used to compute the energy (avoidance) function.
- (void)setEnergyFunction:(NSString *)pipepart;
{
    [pipeline release];
    pipeline = [[NSString alloc] initWithFormat:@"tifftopnm | %@ | ppmtopgm", pipepart];
}

// Pipe 'baseImage' through a sequence of netpbm operations, and read the resulting graymap into the energyFunction[] array (also setting the rows/cols variables and informing the delegate via -setEnergyFuncImage:).
- (void)computeEnergyFunction;
{
    NSMutableData *pipel = [[[pipeline dataUsingEncoding:NSUTF8StringEncoding] mutableCopy] autorelease];
    [pipel appendBytes:"" length:1];  // Append a NUL so that this is usable as a C string
    
    const char *extraPath = "/opt/local/bin";
    
    FILE *tifffp = tmpfile();
    NSData *tiffr = [baseImage TIFFRepresentation];
    fwrite([tiffr bytes], 1, [tiffr length], tifffp);
    fflush(tifffp);
    rewind(tifffp);
    
    int pipeout[2];
    if(pipe(pipeout)) perror("pipe<pipeout>");
    
    pid_t p = fork();
    if (p == 0) {
        dup2(fileno(tifffp), 0);
        lseek(0, 0, SEEK_SET);
        close(pipeout[0]);
        dup2(pipeout[1], 1);
        
        if (extraPath) {
            const char *oldpath = getenv("PATH");
            char *newpath = NULL;
            asprintf(&newpath, "%s:%s", oldpath, extraPath);
            setenv("PATH", newpath, 1);
        }
        
        execl("/bin/sh", "/bin/sh", "-c", [pipel bytes], NULL);
        perror("exec");
        exit(0);
    } else {
        close(pipeout[1]);
    }
    
    fclose(tifffp);
    
    FILE *infp = fdopen(pipeout[0], "r");
    int first = fgetc(infp);
    if (first != 'P') {
        NSRunAlertPanel(@"Command failed",
                        @"Unable to compute the energy function for that image. Check the system console for messages, maybe.",
                        NULL, NULL, NULL);
        return;
    } else {
        ungetc(first, infp);
    }
    int format;
    gray maxval;
    pgm_readpgminit(infp, &cols, &rows, &maxval, &format);
    
    if (energyFunction)
        free(energyFunction);
    energyFunction = malloc(cols*rows*sizeof(*energyFunction));
    
    gray *onerow = pgm_allocrow(cols);
    for(int row = 0; row < rows; row ++) {
        pgm_readpgmrow(infp, onerow, cols, maxval, format);
        uint8_t *outbase = energyFunction + ( row * cols );
        if (maxval == 255) {
            for(int col = 0; col < cols; col ++)
                outbase[col] = onerow[col];
        } else {
            float conv = 256. / ((float)(maxval+1));
            for(int col = 0; col < cols; col ++) {
                outbase[col] = (unsigned int)( ((float)(onerow[col])) * conv );
            }
        }
    }
    pgm_freerow(onerow);
    fclose(infp);
    
    NSBitmapImageRep *outrep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:&energyFunction
                                                                       pixelsWide:cols pixelsHigh:rows
                                                                    bitsPerSample:8 samplesPerPixel:1
                                                                         hasAlpha:NO
                                                                         isPlanar:NO
                                                                   colorSpaceName:NSCalibratedWhiteColorSpace
                                                                     bitmapFormat:0
                                                                      bytesPerRow:cols bitsPerPixel:8];
    NSImage *outimg = [[NSImage alloc] initWithSize:(NSSize){cols, rows}];
    [outimg addRepresentation:outrep];
    [outrep release];
    
    [nonretained_delegate setEnergyFuncImage:outimg];
    [outimg release];
}

// Comparison functions for qsort()ing various values

static int compareIndirect(void *context, const void *l, const void *r)
{
    int *costs = context;
    int lcost = costs[*(short *)l];
    int rcost = costs[*(short *)r];
    
    if (lcost < rcost)
        return -1;
    else if (lcost > rcost)
        return 1;
    else
        return 0;
}

static int compareDirect(const void *l, const void *r)
{
    short ll = *(short *)l;
    short rr = *(short *)r;
    
    if (ll < rr)
        return -1;
    else if (ll > rr)
        return 1;
    else
        return 0;
}

// Once a seam has been computed, record it in the seams[] array and inform the delegate.
// Takes ownership of the colIndices buffer.
- (void)recordSeam:(int *)colIndices
{
    seams = realloc(seams, sizeof(*seams) * ++seamcount);
    seams[seamcount-1] = colIndices;
    [nonretained_delegate anotherSeam];
}

// Returns a sorted list of the columns of 'row' which are to be removed by the first 'skipcount' seams. The returned array size is always equal to skipcount.
// Caller is responsible for not calling this method with skipcount>seamcount.
// Caller must free the returned buffer.
- (short *)makeSkipList:(int)row count:(int)skipcount
{
    short *smap = malloc(sizeof(*smap) * skipcount);
    for(int seam = 0; seam < skipcount; seam ++)
        smap[seam] = seams[seam][row];
    qsort(smap, skipcount, sizeof(*smap), compareDirect);
    return smap;
}

// Returns an array which maps "reduced" columns (see below) to the columns of the original image, if all the seams computed so far have been removed.
// Caller must free the returned buffer.
- (short *)makeSkipMap:(int)row
{
    short *skipmap = malloc(sizeof(*skipmap) * (cols - seamcount));
    short *smap = [self makeSkipList:row count:seamcount];
    
    int reducedCol, imageCol, smapIndex;
    imageCol = 0;
    reducedCol = 0;
    smapIndex = 0;
    while(reducedCol < (cols-seamcount)) {
        while(smapIndex < seamcount && smap[smapIndex] == imageCol) {
            smapIndex ++;
            imageCol ++;
        }
        skipmap[reducedCol] = imageCol;
        reducedCol ++;
        imageCol ++;
    }
    free(smap);
    return skipmap;
}

/*
 -computeSeams is the core of the program (the dynamic-programming seam-finding algorithm).
 
 This is a typical dynamic programming routine. The algorithm proceeds row-by-row from row 0 to the bottom of the image, finding the least-expensive path from the top of the image to that row (where "cost" is the sum of the gray values of the pixels traversed, possibly with some adjustments). upmap[][] holds, for each pixel, the column index of the pixel on the previous line which lies along the least-cost path. thiscost[] and upcost[] hold the minimum costs known for reaching a pixel on the current resp. revious rows. Each iteration thrugh the outermost loop fills in thiscost[] and another row of upmap[][], computing them from upcost[] and the previous row of upmap[][]. The costs are discarded except for the most recent rows, but upmap[][] is retained so the least-cost path can be reconstructed.
 
 There is one complication. Notionally, each invocation of this algorithm is run against an image which has had all previously-found seams carved out (I'll call this the "reduced image"). Rather than do that, this method uses -makeSkipMap: to generate an array that maps columns in the reduced image ("reduced columns") to the column indices of the corresponding pixels in the original image ("full columns"). Note that the expandmap is different for every row. In the below code, some column variables, array indices, and array values are "reduced" column indices, and some are "full" column indices; the comments should hopefully make it clear which is which.
 
*/

#define DIAG_COST 0  // excess cost for diagonal seams
- (void)computeSeams
{
    int reducedCols = cols - seamcount;
    
    short **upmap;  // maps [row][fullcol] to the cheapest reachable fullcol on row-1
    int *upcost,    // maps fullcol to the least cost we can find on that col, for previous row
        *thiscost;  // maps reducedcol to the least cost, for this row
    
    
    upmap = malloc(sizeof(*upmap) * rows);
    for(int row = 0; row < rows; row ++)
        upmap[row] = NULL;
    upcost = malloc(sizeof(*upcost) * cols);
    thiscost = malloc(sizeof(*thiscost) * reducedCols);
    for(int col = 0; col < cols; col ++)
        upcost[col] = 0;
    
    /* Run through the image, findinng the least-cost paths and their costs. */
    
    short *last_expandmap;  // maps reduced col to full col for previous row
    short *this_expandmap;  // maps reduced col to full col for this row
    last_expandmap = NULL;
    for(int row = 0; row < rows; row ++) {
        short *up = malloc(sizeof(*up) * cols);
        upmap[row] = up;
        for(int col = 0; col < cols; col ++)
            up[col] = -2;
        this_expandmap = [self makeSkipMap:row];
        for(int col = 0; col < reducedCols; col ++) {
            int pixelcost = energyFunction[ row*cols + this_expandmap[col] ];
            if (row == 0) {
                thiscost[col] = pixelcost;
                up[this_expandmap[col]] = -2;
                continue;
            }
            int c0 = pixelcost + upcost[last_expandmap[col]];
            int cup, cleft, cright;
            cup = c0;
            int up0 = col;
            if (col > 0) {
                cleft = pixelcost + upcost[last_expandmap[col-1]] + DIAG_COST;
                if (cleft < c0) {
                    c0 = cleft;
                    up0 = col-1;
                }
            }
            if (col < (reducedCols-1)) {
                cright = pixelcost + upcost[last_expandmap[col+1]] + DIAG_COST;
                if (cright  < c0) {
                    c0 = cright;
                    up0 = col+1;
                }
            }
            /* printf("row %d col %3d: costs [%d %d %d] up=%d / %d\n",
                   row, col, cleft, cup, cright, up0, c0); */
            thiscost[col] = c0;
            assert(this_expandmap[col] >= 0);
            assert(this_expandmap[col] < cols);
            up[this_expandmap[col]] = last_expandmap[up0];
        }
        for(int col = 0; col < reducedCols; col ++)
            upcost[this_expandmap[col]] = thiscost[col];
        free(last_expandmap);
        last_expandmap = this_expandmap;
        this_expandmap = NULL;
    }
    
    /* Generate a list of columns mincol[], sorted according to the cost of the path that terminates there (from thiscost[]) */
    
    short *mincol = malloc(sizeof(*mincol) * reducedCols);
    for(int col = 0; col < reducedCols; col ++) {
        mincol[col] = col;
    }
    qsort_r(mincol, reducedCols, sizeof(*mincol), thiscost, compareIndirect);  /* sort according to cost */
    for(int col = 0; col < reducedCols; col ++)
        mincol[col] = last_expandmap[mincol[col]];  /* convert mincol[] from reduced to full column indices */

    free(last_expandmap);

    /* Trace back along the least-cost path, storing the resulting seam's columns in seamcol[] */
    /* We try to find multiple seams here without rerunning the upmap[][] calculation, but we can only do that consistently if they dont intersect (otherwise the costs of the later seam would ahve been affected by the carving-out of the previous seam). Unfortunately we rarely get more than one seam per run for most images. */
    
    int *seamcols = NULL;
    for(int seam = 0; seam < reducedCols; seam ++) {
        int col = mincol[seam];
        int row;
        if (!seamcols)
            seamcols = malloc(sizeof(*seamcols) * rows);
        for(row = rows-1; row >= 0; ) {
            assert(col >= 0 && col < cols);
            seamcols[row] = col;
            short nextcol = upmap[row][col];
            upmap[row][col] = -1;
            if (nextcol == -1)
                break;
            col = nextcol;
            row --;
        }
        if (row < 0) {
            [self recordSeam:seamcols];
            seamcols = NULL;
        } else {
            break;
        }
    }
    if(seamcols)
        free(seamcols);
    
    free(mincol);
    free(upcost);
    free(thiscost);
    
    {
        /* Store the upmap in the instance variable so the seam view can use it; swap the old copy into upmap[][] so we can deallocate its memory. */
        short **t = last_upmap;
        last_upmap = upmap;
        upmap = t;
    }
    if (upmap) {
        for(int row = 0; row < rows; row ++)
            free(upmap[row]);
        free(upmap);
    }
}

- (void)doComputeSeams:sender;  // a more-or-less redundant method
{
    [self computeSeams];
}

@end
